"""Loading/dumping context holding the current execution stack, calling serializers and validators."""
from __future__ import annotations

__all__ = ['Context', 'Loading', 'Dumping']

from abc import ABC, abstractmethod
from collections import deque
from typing import Any, TypeVar, Deque

from serious.serialization.serializer import Serializer
from serious.types import FrozenList
from serious.validation import validate

M = TypeVar('M')  # Python model value
S = TypeVar('S')  # Serialized value

SerializationStep = str

class Context(ABC):
    """An abstract base class for the serialization context.

    Serialization context is created when executing `SeriousModel` load or dump and is passed inside
    all nested serializers.

    All of serializers are called via context to include them in stack and to perform
    all the necessary validation and processing
    """
    _steps: Deque[SerializationStep]
    validating: bool
    _last_validated_value: Any

    def __init__(self, steps: Deque[SerializationStep], validating: bool):
        self._steps = steps
        self.validating = validating
        self._last_validated_value = None

    @property
    def path(self):
        return ''.join(self._steps)

    @property
    def stack(self) -> FrozenList[SerializationStep]:
        """The stack is included in errors, mentioning the fields, array indexes, dictionary keys, etc."""
        return FrozenList(self._steps)

    def __repr__(self):
        return f"<Context: {self.path}>"

    @abstractmethod
    def run(self, step: str, serializer: Serializer, value: Any) -> Any:
        """Execute serializer in context.

        Implementations:
        - includes the current step in the stack,
        - executes current steps serializer,
        - performs any required processing of values.

        This abstraction is needed for a straightforward custom serializer implementation.
        Extracting validation (or any other value processing) from serializers to `Context#run(...)`
        has left field serializers with a plain structure of a constructor, load, dump, and fits methods.
        """
        raise NotImplementedError

    def validate(self, o):
        self._steps.append(f".__validate__()")
        self._last_validated_value = o
        validate(o)
        self._steps.pop()

    def _repr_last_validated_value(self):
        value_repr = repr(self._last_validated_value)
        if len(value_repr) > 50:
            return f'{value_repr[:50]}<...+{len(value_repr) - 50}>'
        return value_repr

    def failed_validation_at(self):
        value = self._repr_last_validated_value()
        return f"Failed validation of {value} at \"{self.path}\""

class Loading(Context):
    """Context used during **load** operations."""

    def __init__(self, *, validating: bool, root: str = '__root__', steps: Deque[SerializationStep] | None = None):
        if steps is None:
            steps = deque()
            steps.append(root)
        super().__init__(steps, validating)

    def run(self, step: str, serializer: Serializer[M, S], value: S) -> M:
        self._steps.append(step)
        self._last_validated_value = value
        result = serializer.load(value, self)
        if self.validating:
            self.validate(result)
        self._steps.pop()
        return result



class Dumping(Context):
    """Context used during **dump** operations."""

    def __init__(self, *, validating: bool, root: str = '.', steps: Deque[SerializationStep] | None = None):
        if steps is None:
            steps = deque()
            steps.append(root)
        super().__init__(steps, validating)

    def run(self, step: str, serializer: Serializer[M, S], o: M) -> S:
        self._steps.append(step)
        if self.validating:
            self.validate(o)
        self._last_validated_value = o
        result = serializer.dump(o, self)
        self._steps.pop()
        return result

    def validation_proxy(self) -> Loading:
        return DumpingValidationProxy(self)


class DumpingValidationProxy(Loading):
    def __init__(self, ctx: Dumping):
        super().__init__(validating=False)
        self._ctx = ctx

    @property
    def _steps(self):
        return self._ctx._steps

    @_steps.setter
    def _steps(self, value):
        pass

    @property
    def _last_validated_value(self):
        return self._ctx._last_validated_value

    @_last_validated_value.setter
    def _last_validated_value(self, value):
        pass



from __future__ import annotations

__all__ = [
    'field_serializers',
    'OptionalSerializer', 'AnySerializer', 'EnumSerializer', 'DictSerializer', 'CollectionSerializer',
    'TupleSerializer', 'StringSerializer', 'BooleanSerializer', 'IntegerSerializer', 'FloatSerializer',
    'DataclassSerializer', 'UtcTimestampSerializer', 'DateTimeIsoSerializer', 'DateIsoSerializer',
    'TimeIsoSerializer', 'UuidSerializer', 'DecimalSerializer'
]

import re
from dataclasses import replace
from datetime import datetime, date, time
from decimal import Decimal
from enum import Enum
from types import UnionType
from typing import Any, Optional, Dict, List, Union, Pattern, Iterable, Type, Tuple, Literal, TypeVar
from uuid import UUID

from serious.descriptors import TypeDescriptor
from serious.errors import ValidationError
from serious.types import Timestamp, FrozenList, FrozenDict
from .context import Context, Loading, Dumping
from .serializer import FieldSerializer, Serializer


def field_serializers(custom: Iterable[Type[FieldSerializer]] = tuple()) -> Tuple[Type[FieldSerializer], ...]:
    """Default list of Serious field serializers.

    Returns a frozen collection of field serializers in the default order.
    You can provide a list of custom field serializers to include them along with default serializers.
    The order in the collection defines the order in which the serializers will be tested for fitness for each field.

    :param custom: a list of custom serializers which are injected into the default list

    """

    extras: List[Type[FieldSerializer]] = []

    if SQLALCHEMY_INTEGRATION_ENABLED:
        extras.append(SqlAlchemyDeclarativeSerializer)

    if PYDANTIC_INTEGRATION_ENABLED:
        extras.append(PydanticModelSerializer)

    return tuple([
        OptionalSerializer,
        UnionSerializer,
        AnySerializer,
        LiteralSerializer,
        EnumSerializer,
        *custom,
        TypedDictSerializer,
        DictSerializer,
        CollectionSerializer,
        TupleSerializer,
        StringSerializer,
        BooleanSerializer,
        IntegerSerializer,
        FloatSerializer,
        DataclassSerializer,
        UtcTimestampSerializer,
        DateTimeIsoSerializer,
        DateIsoSerializer,
        TimeIsoSerializer,
        UuidSerializer,
        DecimalSerializer,
        *extras,
    ])


class OptionalSerializer(FieldSerializer[Optional[Any], Optional[Any]]):
    """
    A serializer for field marked as `Optional`. An optional field has internally a serializer for the target type,
    but first checks if the loaded data is `None`.

        :Example:

        @dataclass
        class Node:
            node: Optional[Node]
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        item_descriptor = replace(self.type, is_optional=False)
        self._serializer = self.root.find_serializer(item_descriptor)

    def load(self, value: Optional[Any], ctx: Loading) -> Optional[Any]:
        return None if value is None else self._serializer.load(value, ctx)

    def dump(self, value: Optional[Any], ctx: Dumping) -> Optional[Any]:
        return None if value is None else self._serializer.dump(value, ctx)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return desc.is_optional


class UnionSerializer(FieldSerializer[Any, Dict]):
    """
    A serializer for Union fields.

        :Example:

        @dataclass
        class Character:
            weapon: Union[Sword, Staff, Hammer]
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._serializers_by_cls = {
            desc.cls: self.root.find_serializer(desc) for desc in self.type.parameters.values()
        }
        self._serializers_by_name = {cls.__name__: serializer for cls, serializer in self._serializers_by_cls.items()}

    def load(self, value: Dict, ctx: Loading) -> Any:
        try:
            value = dict(value)
        except TypeError:
            raise ValidationError(f'Invalid Union[{",".join(self._serializers_by_name)}] value: {value}, '
                                  f'must be a dict with "__type__" and "__value__" keys')

        try:
            t = value['__type__']
        except KeyError:
            raise ValidationError(f'Invalid Union[{",".join(self._serializers_by_name)}] value: {value}, '
                                  f'missing "__type__" key')
        try:
            v = value['__value__']
        except KeyError:
            raise ValidationError(f'Invalid Union[{",".join(self._serializers_by_name)}] value: {value}, '
                                  f'missing "__value__" key')
        try:
            serializer = self._serializers_by_name[t]
        except KeyError:
            raise ValidationError(f'Invalid Union[{",".join(self._serializers_by_name)}] value: {value}, '
                                  f'unsupported type.')
        return serializer.load(v, ctx)

    def dump(self, value: Any, ctx: Dumping) -> Dict:
        try:
            serializer = self._serializers_by_cls[type(value)]
        except KeyError:
            raise ValidationError(f'Invalid Union[{",".join(self._serializers_by_name)}] value: {value}')
        return {
            '__type__': type(value).__name__,
            '__value__': serializer.dump(value, ctx),
        }

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return desc.cls is Union or isinstance(desc.cls, UnionType)


class EnumSerializer(FieldSerializer[Any, Any]):
    """Enum value serializer. Note that output depends on enum value, so it can be `str`, `int`, etc.

    It is possible to serialize enums of non-S type if the enum is supplying this type as parent class.
    For example a date serialized to ISO string:

        :Example:

        class Date(date, Enum):
            TRINITY = 1945, 6, 16
            GAGARIN = 1961, 4, 11


        @dataclass(frozen=True)
        class HistoricEvent:
            name: str
            date: Date

        model = DictModel(HistoricEvent)
        dict = {'name': name, 'date': '1961-04-11'}
        dataclass = HistoricEvent(name, Date.GAGARIN)
        assert model.load(dict) == dataclass  # True
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._serializer = self._value_serializer()
        self._enum_values = {e.name for e in list(self.type.cls)}

    def _value_serializer(self) -> Optional[FieldSerializer]:
        cls = self.type.cls
        bases = cls.__bases__
        while len(bases) == 1:
            cls = bases[0]
            bases = cls.__bases__
        if len(bases) == 0:
            return None
        item_descriptor = self.type.describe(bases[0])
        return self.root.find_serializer(item_descriptor)

    def load(self, value: Any, ctx: Loading) -> Any:
        if self._serializer is None and value not in self._enum_values:
            raise ValidationError(f'"{value}" is not part of the {self.type.cls} enum')
        enum_cls = self.type.cls
        if self._serializer is not None:
            loaded_value = self._serializer.load(value, ctx)
            try:
                return enum_cls(loaded_value)
            except ValueError as e:
                raise ValidationError(f'"{value}" is not part of the {enum_cls} enum') from e
        return enum_cls[value]

    def dump(self, value: Any, ctx: Dumping) -> Any:
        if self._serializer is not None:
            return self._serializer.dump(value.value, ctx)
        return value.name

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, Enum)


class AnySerializer(FieldSerializer[Any, Any]):
    """Serializer for `Any` fields."""

    def load(self, value: Any, ctx: Loading) -> Any:
        return value

    def dump(self, value: Any, ctx: Dumping) -> Any:
        return value

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return desc.cls is Any


class LiteralSerializer(FieldSerializer[Any, Any]):
    """Serializer for `Any` fields."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dump_values = {
            v.cls: self.root.find_serializer(self.type.describe(v.cls.__class__)).dump(v, Dumping(validating=False, root='root'))
            for v in self.type.parameters.values()
        }
        self._load_values = {
            value: key
            for key, value in self._dump_values.items()
        }

    def load(self, value: Any, ctx: Loading) -> Any:
        return value

    def dump(self, value: Any, ctx: Dumping) -> Any:
        return value

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return desc.cls is Literal


class TypedDictSerializer(FieldSerializer[Dict[str, Any], Dict[str, Any]]):
    """Serializer for `TypedDict` fields."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._field_serializers = {}
        for field, desc in self.type.fields.items():
            self._field_serializers[field] = Alias(self.root.find_serializer(desc))

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return desc.is_typed_dict

    def load(self, data: Dict[str, Any], ctx: Loading) -> Dict[str, Any]:
        if not isinstance(data, dict):
            raise ValidationError('Expecting a dictionary')
        items = self._serialize_typed_dict(data, ctx)
        return self.type.cls(items)

    def dump(self, data: Dict[str, Any], ctx: Dumping) -> Dict[str, Any]:
        return self._serialize_typed_dict(data, ctx)

    def _serialize_typed_dict(self, data: Dict[str, Any], ctx: Context) -> Dict[str, Any]:
        if missing := {field for field in self._field_serializers if field not in data}:
            raise ValidationError(f"Missing fields: {missing}")
        return {
            key: ctx.run(f'[key]', serializer(key), data[key])
            for key, serializer in self._field_serializers.items()
        }


class DictSerializer(FieldSerializer[Dict[str, Any], Dict[str, Any]]):
    """Serializer for `dict` fields with `str` keys (`Dict[str, Any]`)."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        key_desc = self.type.parameters[0]
        value_desc = self.type.parameters[1]
        assert not key_desc.is_optional, 'Dict keys must have explicit "str" type (Dict[str, Any]).'
        self._key_serializer = self.root.find_serializer(key_desc)
        self._value_serializer = self.root.find_serializer(value_desc)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, dict) or issubclass(desc.cls, FrozenDict)

    def load(self, data: Dict[str, Any], ctx: Loading) -> Dict[str, Any]:
        if not isinstance(data, dict):
            raise ValidationError('Expecting a dictionary')
        items = self._serialize_dict(data, ctx)
        return self.type.cls(items)

    def dump(self, data: Dict[str, Any], ctx: Dumping) -> Dict[str, Any]:
        return self._serialize_dict(data, ctx)

    def _serialize_dict(self, data: Dict[str, Any], ctx: Context) -> Dict[str, Any]:
        key_serializer = Alias(self._key_serializer)
        value_serializer = Alias(self._value_serializer)
        return {
            ctx.run(f'#{key}', key_serializer(key), key): ctx.run(f'[{key}]', value_serializer(key), value)
            for key, value in data.items()
        }


Collection = Union[list, set, frozenset]


class CollectionSerializer(FieldSerializer[Collection, list]):
    """Serializer for lists, sets, and frozensets."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._serializer = self.root.find_serializer(self.type.parameters[0])
        self._item_type = self._serializer.type

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return (issubclass(desc.cls, (list, set, frozenset, FrozenList))
                or (issubclass(desc.cls, tuple)
                    and len(desc.parameters) == 2
                    and desc.parameters[1].cls is Ellipsis))

    def load(self, value: list, ctx: Loading) -> Collection:
        if not isinstance(value, list):
            raise ValidationError(f'Expecting a list of {self._item_type.cls} values')
        items = self._serialize_collection(value, ctx)
        return self.type.cls(items)

    def dump(self, value: Collection, ctx: Dumping) -> list:
        return self._serialize_collection(value, ctx)

    def _serialize_collection(self, data: Any, ctx: Context) -> List[Any]:
        serializer = Alias(self._serializer)
        return [ctx.run(f'[{i}]', serializer(i), item) for i, item in enumerate(data)]


class TupleSerializer(FieldSerializer[tuple, list]):
    """Serializer for Python tuples."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._serializers = [self.root.find_serializer(self.type.parameters[i]) for i in self.type.parameters]
        self._size = len(self._serializers)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, tuple)

    def load(self, value: list, ctx: Loading) -> tuple:
        if not isinstance(value, list):
            raise ValidationError(f'Expecting a list of {self._size} tuple values')
        if len(value) != self._size:
            raise ValidationError(f'Expecting a list of {self._size} tuple values')  # type: ignore
        items = self._serialize_tuple(value, ctx)
        return self.type.cls(items)

    def dump(self, value: tuple, ctx: Dumping) -> list:
        return self._serialize_tuple(value, ctx)

    def _serialize_tuple(self, data: Any, ctx: Context) -> List[Any]:
        serializer = OrdinalAlias(self._serializers)
        return [ctx.run(f'[{i}]', serializer(i), item) for i, item in enumerate(data)]


class Alias(Serializer):
    """Serializes values using the constructor parameter, but step name via calling instance as a function."""

    def __init__(self, serializer):
        self._serializer = serializer

    def __call__(self, key: Union[str, int]) -> Alias:
        self._key = key
        return self

    def step_name(self) -> str:
        return f'[{self._key}]'

    def load(self, value: Any, ctx: Loading) -> Any:
        return self._serializer.load(value, ctx)

    def dump(self, value: Any, ctx: Dumping) -> Any:
        return self._serializer.dump(value, ctx)


S = TypeVar("S", bound=Serializer)


class OrdinalAlias(Serializer[Any, Any]):
    """Serializes values using the provided serializer at the currently set index.
    The index is set via calling instance as a function.
    Step name is changed to match the index.
    """

    def __init__(self, serializers: List[S]):
        self._serializers = serializers

    def __call__(self, index: int) -> OrdinalAlias:
        self._index = index
        return self

    def step_name(self) -> str:
        return f'[{self._index}]'

    def load(self, value: Any, ctx: Loading) -> Any:
        return self._serializers[self._index].load(value, ctx)

    def dump(self, value: Any, ctx: Dumping) -> Any:
        return self._serializers[self._index].dump(value, ctx)


class BooleanSerializer(FieldSerializer[bool, bool]):
    """A serializer boolean field values."""

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, bool)

    def load(self, value: bool, ctx: Loading) -> bool:
        if not isinstance(value, bool):
            raise ValidationError(f"Invalid data type. Expecting boolean")
        return self.type.cls(value)

    def dump(self, value: bool, ctx: Dumping) -> bool:
        return bool(value)


class StringSerializer(FieldSerializer[str, str]):
    """A serializer for string field values."""

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, str)

    def load(self, value: str, ctx: Loading) -> str:
        if not isinstance(value, str):
            raise ValidationError('Invalid data type. Expecting a string')
        return self.type.cls(value)

    def dump(self, value: str, ctx: Dumping) -> str:
        return str(value)


class IntegerSerializer(FieldSerializer[int, int]):
    """A serializer for integer field values.

    .. note::
        In Python `bool` is a subclass of `int`, thus the check.
    """

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, int) and not issubclass(desc.cls, bool)

    def load(self, value: int, ctx: Loading) -> int:
        if not isinstance(value, int) or isinstance(value, bool):
            raise ValidationError('Invalid data type. Expecting an integer')
        return self.type.cls(value)

    def dump(self, value: int, ctx: Dumping) -> int:
        return int(value)


class FloatSerializer(FieldSerializer[float, float]):
    """A serializer for float values field values.

    During load this can be either an int (1) or float (1.0). Always dumps to float."""

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, float)

    def load(self, value: float, ctx: Loading) -> float:
        is_numeric = isinstance(value, (int, float)) and not isinstance(value, bool)
        if not is_numeric:
            raise ValidationError('Invalid data type. Expecting a numeric value')
        return self.type.cls(value)

    def dump(self, value: float, ctx: Dumping) -> float:
        return float(value)


class DataclassSerializer(FieldSerializer[Any, Dict[str, Any]]):
    """A serializer for field values that are dataclasses instances."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._serializer = self.root.child_model(self.type)
        self._dc_name = self.type.cls.__name__

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return desc.is_dataclass

    def load(self, value: Dict[str, Any], ctx: Loading) -> Any:
        if not isinstance(value, dict):
            raise ValidationError(f'Invalid data type. Expecting a mapping matching {self._dc_name} model')
        return self._serializer.load(value, ctx)  # type: ignore # type: ignore # value always a mapping

    def dump(self, value: Any, ctx: Dumping) -> Dict[str, Any]:
        return self._serializer.dump(value, ctx)


class UtcTimestampSerializer(FieldSerializer[Timestamp, Union[float, int]]):
    """A serializer of UTC timestamp field values to/from float value.

        :Example:

        from serious.types import timestamp

        @dataclass
        class Transaction:
            created_at: timestamp

        transaction = Transaction(timestamp(1542473728.456753))

    Dumping the `post` will return `{"created_at": 1542473728.456753}`.
    """

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, Timestamp)

    def load(self, value: Union[float, int], ctx: Loading) -> Timestamp:
        if not isinstance(value, (int, float)):
            raise ValidationError('Invalid data type. Expecting int or float')
        return Timestamp(value)  # type: ignore # expecting float

    def dump(self, value: Timestamp, ctx: Dumping) -> float:
        return value.value


_iso_date_time_re = re.compile(  # https://stackoverflow.com/a/43931246/8677389
    r'\A(?:\d{4}-(?:(?:0[1-9]|1[0-2])-(?:0[1-9]|1\d|2[0-8])|(?:0[13-9]|1[0-2])-(?:29|30)|(?:0[13578]|1[02])-31)'
    r'|(?:[1-9]\d(?:0[48]|[2468][048]|[13579][26])|(?:[2468][048]|[13579][26])00)-02-29)'
    r'T(?:[01]\d|2[0-3]):[0-5]\d:[0-5]\d(?:\.\d{1,9})?(?:Z|[+-][01]\d:[0-5]\d)?\Z'
)
_iso_date_re = re.compile(
    r'\A(?P<year>[0-9]{4})(?P<hyphen>-?)(?P<month>1[0-2]|0[1-9])(?P=hyphen)(?P<day>3[01]|0[1-9]|[12][0-9])\Z'
)
_iso_time_re = re.compile(
    r'\A(?P<hour>2[0-3]|[01][0-9])'
    r':?(?P<minute>[0-5][0-9])'
    r':?(?P<second>[0-5][0-9])'
    r'(?P<timezone>Z|[+-](?:2[0-3]|[01][0-9])(?::?(?:[0-5][0-9]))?)?\Z'
)


class DateTimeIsoSerializer(FieldSerializer[datetime, str]):
    """A serializer for datetime field values to a timestamp represented by a `ISO formatted string`_.

        :Example:

        @dataclass
        class Post:
            timestamp: datetime

        timestamp = datetime(2018, 11, 17, 16, 55, 28, 456753, tzinfo=timezone.utc)
        post = Post(timestamp)

    Dumping the `post` will return `'{"timestamp": "2018-11-17T16:55:28.456753+00:00"}'`.

    .. _ISO formatted string: https://en.wikipedia.org/wiki/ISO_8601
    """

    def load(self, value: str, ctx: Loading) -> datetime:
        if not isinstance(value, str):
            raise ValidationError('Invalid data type. Expecting a string')
        if not _matches(_iso_date_time_re, value):
            raise ValidationError('Invalid date/time format. Check the ISO 8601 specification')
        return datetime.fromisoformat(value)  # type: ignore # expecting datetime

    def dump(self, value: datetime, ctx: Dumping) -> str:
        return datetime.isoformat(value)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, datetime)


class DateIsoSerializer(FieldSerializer[date, str]):
    """A serializer of `date` field values to a timestamp represented by an `ISO formatted string`_.

        :Example:

        @dataclass
        class Event:
            name: str
            when: date

        event = Event('Albert Einstein won Nobel Prize in Physics', date(1922, 9, 9))

    Dumping the `event` will return `'{"name": "…", "when": "1922-09-09"}'`.

    .. _ISO formatted string: https://en.wikipedia.org/wiki/ISO_8601
    """

    def load(self, value: str, ctx: Loading) -> date:
        if not isinstance(value, str):
            raise ValidationError('Invalid data type. Expecting a string')
        if not _matches(_iso_date_re, value):
            raise ValidationError('Invalid date format. Check the ISO 8601 specification')
        return date.fromisoformat(value)  # type: ignore # expecting datetime

    def dump(self, value: date, ctx: Dumping) -> str:
        return date.isoformat(value)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, date)


class TimeIsoSerializer(FieldSerializer[time, str]):
    """A serializer for `time` field values to an `ISO formatted string`_.

        :Example:

        @dataclass
        class Alarm:
            at: time
            enabled: bool

        alarm = Alarm(time(7, 0, 0), enabled=True)

    Dumping the `post` will return `'{"at": "07:00:00", "enabled": True}'`.

    .. _ISO formatted string: https://en.wikipedia.org/wiki/ISO_8601
    """

    def load(self, value: str, ctx: Loading) -> time:
        if not isinstance(value, str):
            raise ValidationError('Invalid data type. Expecting a string')
        if not _matches(_iso_time_re, value):
            raise ValidationError('Invalid time format. Check the ISO 8601 specification')
        return time.fromisoformat(value)  # type: ignore # expecting datetime

    def dump(self, value: time, ctx: Dumping) -> str:
        return time.isoformat(value)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, time)


_uuid_hex_re = re.compile(r'\A([a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{12})\Z', re.I)


class UuidSerializer(FieldSerializer[UUID, str]):
    """A `UUID` value serializer to `str`."""

    def load(self, value: str, ctx: Loading) -> UUID:
        if not isinstance(value, str):
            raise ValidationError('Invalid data type. Expecting a string')
        if not _matches(_uuid_hex_re, value):
            raise ValidationError('Invalid UUID hex format')
        return UUID(value)  # type: ignore # expecting str

    def dump(self, value: UUID, ctx: Dumping) -> str:
        return str(value)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, UUID)


_decimal_re = re.compile(r'\A\d+(\.\d+)?\Z')


class DecimalSerializer(FieldSerializer[Decimal, str]):
    """`Decimal` value serializer to `str`."""

    def load(self, value: str, ctx: Loading) -> Decimal:
        if not isinstance(value, str):
            raise ValidationError('Invalid data type. Expecting a string')
        if not _matches(_decimal_re, value):
            raise ValidationError('Invalid decimal format. A number with a "." as a decimal separator is expected')
        return Decimal(value)  # type: ignore # expecting str

    def dump(self, value: Decimal, ctx: Dumping) -> str:
        return str(value)

    @classmethod
    def fits(cls, desc: TypeDescriptor) -> bool:
        return issubclass(desc.cls, Decimal)


def _matches(regex: Pattern, value: str) -> bool:
    return regex.match(value) is not None  # type: ignore # caller ensures str


try:
    from sqlalchemy.orm import DeclarativeMeta

    SQLALCHEMY_INTEGRATION_ENABLED = True

    class SqlAlchemyDeclarativeSerializer(
        FieldSerializer[DeclarativeMeta, Dict[str, Any]]
    ):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._field_serializers = {}
            for field, desc in self.type.fields.items():
                if not (desc.is_sqlalchemy_model or any(p.is_sqlalchemy_model for p in desc.parameters.values())):
                    self._field_serializers[field] = Alias(self.root.find_serializer(desc))

        @classmethod
        def fits(cls, desc: TypeDescriptor) -> bool:
            return desc.is_sqlalchemy_model

        def load(self, data: Dict[str, Any], ctx: Loading) -> DeclarativeMeta:
            if not isinstance(data, dict):
                raise ValidationError("Expecting a dictionary")
            if missing := {field for field in self._field_serializers if field not in data}:
                raise ValidationError(f"Missing fields: {missing}")
            items = {
                key: ctx.run(f".{key}", serializer(key), data[key])
                for key, serializer in self._field_serializers.items()
            }
            return self.type.cls(**items)

        def dump(self, data: DeclarativeMeta, ctx: Dumping) -> Dict[str, Any]:
            return {
                key: ctx.run(f".{key}", serializer(key), getattr(data, key))
                for key, serializer in self._field_serializers.items()
            }

except ImportError:
    SQLALCHEMY_INTEGRATION_ENABLED = False

try:
    from pydantic import BaseModel

    PYDANTIC_INTEGRATION_ENABLED = True


    class PydanticModelSerializer(FieldSerializer[BaseModel, str]):
        @classmethod
        def fits(cls, desc: TypeDescriptor) -> bool:
            return issubclass(desc.cls, BaseModel)

        def dump(self, value: BaseModel, ctx: Dumping) -> str:
            return value.json()

        def load(self, value: str, ctx: Loading) -> BaseModel:
            return self.type.cls.parse_raw(value)

except ImportError:
    PYDANTIC_INTEGRATION_ENABLED = False



"""Errors specific to JSON model."""


class UnexpectedJson(Exception):
    """Invalid JSON provided to JSON model."""

    def __init__(self, extra_message=None):
        super().__init__(f'Unexpected JSON document. {extra_message}')



"""Serialization module includes implementation of Serious internal model (`SeriousModel`),
serializers (`FieldSerializer` and subclasses), and more (listed in `__all__`).

You’ll be referring to contents of this model when creating custom field serializers.
"""

from __future__ import annotations

__all__ = [
    'SeriousModel',
    'FieldSerializer',
    'field_serializers',
    'Loading',
    'Dumping',
    'KeyMapper',
    'OptionalSerializer',
    'AnySerializer',
    'EnumSerializer',
    'DictSerializer',
    'CollectionSerializer',
    'TupleSerializer',
    'StringSerializer',
    'BooleanSerializer',
    'IntegerSerializer',
    'FloatSerializer',
    'DataclassSerializer',
    'UtcTimestampSerializer',
    'DateTimeIsoSerializer',
    'DateIsoSerializer',
    'TimeIsoSerializer',
    'UuidSerializer',
    'DecimalSerializer',
]

from .context import Loading, Dumping, SerializationStep
from .field_serializers import field_serializers, OptionalSerializer, AnySerializer, EnumSerializer, DictSerializer, \
    CollectionSerializer, TupleSerializer, StringSerializer, BooleanSerializer, IntegerSerializer, FloatSerializer, \
    DataclassSerializer, UtcTimestampSerializer, DateTimeIsoSerializer, DateIsoSerializer, TimeIsoSerializer, \
    UuidSerializer, DecimalSerializer
from .model import SeriousModel
from .key_mapper import KeyMapper
from .serializer import Serializer, FieldSerializer



"""
`SeriousModel` is the core of Serious implementation traversing the dataclass, checking it’s correctness,
 and serializing it to/from dictionaries of primitives.
"""
from __future__ import annotations

__all__ = ['SeriousModel']

from dataclasses import fields, MISSING, Field, is_dataclass
from typing import Generic, Iterable, Type, Dict, Any, Union, Mapping, Optional, Iterator, TypeVar

from serious.checks import check_is_instance
from serious.descriptors import scan_types, TypeDescriptor
from serious.errors import ModelContainsAny, MissingField, UnexpectedItem, ValidationError, \
    LoadError, DumpError, FieldMissingSerializer
from serious.utils import Dataclass
from .check_immutable import check_immutable
from .context import Loading, Dumping
from .key_mapper import KeyMapper, NoopKeyMapper
from .serializer import FieldSerializer

T = TypeVar('T')
M = TypeVar('M')  # Python model value
S = TypeVar('S')  # Serialized value


class SeriousModel(Generic[T]):
    """Serious internal model implementation reused by the exposed models (like JSON/YAML/dict/etc).

    Checks the provided descriptor for model errors when instantiated.

    Used to serialize dataclasses to python dictionaries of primitives.
    Dictionaries of primitives are then transformed to target output formats like JSON by other tools.
    For JSON it’s Python built-in `json` module.
    """

    def __init__(
            self,
            descriptor: TypeDescriptor,
            serializers: Iterable[Type[FieldSerializer]],
            *,
            allow_any: bool,
            allow_missing: bool,
            allow_unexpected: bool,
            validate_on_load: bool,
            validate_on_dump: bool,
            ensure_frozen: Union[bool, Iterable[Type]],
            key_mapper: Optional[KeyMapper] = None,
            _registry: Optional[Dict[TypeDescriptor, SeriousModel]] = None
    ):
        """Initialize a Serious Model.

        :param descriptor: the descriptor of the dataclass to load/dump.
        :param serializers: field serializer classes in an order they will be tested for fitness for each field.
        :param allow_any: `False` to raise if the model contains fields annotated with `Any`
                    (this includes generics like `List[Any]`, or simply `list`).
        :param allow_missing: `False` to raise during load if data is missing the optional fields.
        :param allow_unexpected: `False` to raise during load if data contains some unknown fields.
        :param validate_on_load: to call dataclass __validate__ method after object construction.
        :param validate_on_load: to call object __validate__ before dumping.
        :param ensure_frozen: `False` to skip check of model immutability; `True` will perform the check
                against built-in immutable types; a list of custom immutable types is added to built-ins.
        :param key_mapper: remap field names of between dataclass and serialized objects.
        :param _registry: a mapping of dataclass type descriptors to corresponding serious serializer;
                used internally to create child serializers.
        """
        assert is_dataclass(descriptor.cls), 'Serious can only operate on dataclasses.'
        all_types = scan_types(descriptor)
        if not allow_any and Any in all_types:
            raise ModelContainsAny(descriptor.cls)
        if ensure_frozen:
            check_immutable(descriptor, all_types, ensure_frozen)
        self.descriptor = descriptor
        self.serializers = tuple(serializers)
        self.allow_any = allow_any
        self.allow_missing = allow_missing
        self.allow_unexpected = allow_unexpected
        self.validate_on_load = validate_on_load
        self.validate_on_dump = validate_on_dump
        self.ensure_frozen = ensure_frozen
        self.serializer_registry = {descriptor: self} if not _registry else _registry
        self.keys = key_mapper or NoopKeyMapper()
        self.serializers_by_field = {name: self.find_serializer(desc) for name, desc in descriptor.fields.items()}

    @property
    def cls(self) -> Type[T]:
        # A shortcut to root dataclass type.
        return self.descriptor.cls

    def load(self, data: Mapping, _ctx: Optional[Loading] = None) -> T:
        """Loads dataclass from a dictionary or other mapping. """

        check_is_instance(data, Mapping, f'Invalid data for {self.cls}')  # type: ignore
        root = _ctx is None
        loading: Loading = Loading(
            validating=self.validate_on_load,
            root=self.cls.__name__,
        ) if root else _ctx  # type: ignore # checked above
        mut_data = {self.keys.to_model(key): value for key, value in data.items()}
        if self.allow_missing:
            for field in fields_missing_from(mut_data, self.cls):
                mut_data[field.name] = None
        else:
            check_for_missing(self.cls, mut_data)
        if not self.allow_unexpected:
            check_for_unexpected(self.cls, mut_data)
        try:
            init_kwargs = {
                field: loading.run(f'.{self.keys.to_serialized(field)}', serializer, mut_data[field])
                for field, serializer in self.serializers_by_field.items()
                if field in mut_data
            }
            result = self.cls(**init_kwargs)  # type: ignore # not an object
            if self.validate_on_load:
                loading.validate(result)
            return result
        except ValidationError as e:
            if root:
                raise ValidationError(f"{loading.failed_validation_at()}: {e}") from e
            raise
        except Exception as e:
            if root:
                raise LoadError(self.cls, loading.stack, data) from e
            raise

    def dump(self, o: T, _ctx: Optional[Dumping] = None) -> Dict[str, Any]:
        """Dumps a dataclass object to a dictionary."""

        check_is_instance(o, self.cls)
        root = _ctx is None
        dumping: Dumping = Dumping(
            validating=False,
            root=self.cls.__name__,
        ) if root else _ctx  # type: ignore # checked above
        try:
            _s = self.keys.to_serialized
            result = {
                _s(field): dumping.run(f'.{_s(field)}', serializer, getattr(o, field))
                for field, serializer in self.serializers_by_field.items()
            }
            if self.validate_on_dump:
                dumping.validate(self.load(result, dumping.validation_proxy()))
            return result
        except ValidationError as e:
            if root:
                raise ValidationError(f"{dumping.failed_validation_at()}: {e}") from e
            raise
        except Exception as e:
            if root:
                raise DumpError(o, dumping.stack) from e
            raise

    def child_model(self, descriptor: TypeDescriptor) -> SeriousModel:
        """
        Creates a `SeriousModel` for dataclass fields nested in the current serializers.
        The preferences of the nested dataclasses match those of the root one.
        """
        if descriptor == self.descriptor:
            return self
        if descriptor in self.serializer_registry:
            return self.serializer_registry[descriptor]
        new_model: SeriousModel = SeriousModel(
            descriptor=descriptor,
            serializers=self.serializers,
            allow_any=self.allow_any,
            allow_missing=self.allow_missing,
            allow_unexpected=self.allow_unexpected,
            validate_on_load=self.validate_on_load,
            validate_on_dump=self.validate_on_dump,
            ensure_frozen=self.ensure_frozen,
            key_mapper=self.keys,
            _registry=self.serializer_registry
        )
        self.serializer_registry[descriptor] = new_model
        return new_model

    def find_serializer(self, descriptor: TypeDescriptor) -> FieldSerializer:
        """
        Creates a serializer fitting the provided field descriptor.

        :param descriptor: descriptor of a field to serialize.
        """
        for serializer in self.serializers:
            if serializer.fits(descriptor):
                return serializer(descriptor, self)
        raise FieldMissingSerializer(self.descriptor.cls, descriptor)


def check_for_missing(cls: Type[Dataclass], data: Mapping) -> None:
    """ Checks for missing keys in data that are part of the provided dataclass.
    :raises: MissingField
    """
    missing_fields = filter(lambda f: f.name not in data, fields(cls))
    first_missing_field: Any = next(missing_fields, MISSING)
    if first_missing_field is not MISSING:
        field_names = {first_missing_field.name} | {field.name for field in missing_fields}
        raise MissingField(cls, data, field_names)


def check_for_unexpected(cls: Type[Dataclass], data: Mapping) -> None:
    """ Checks for keys in data that are not part of the provided dataclass.
    :raises: UnexpectedItem
    """
    field_names = {field.name for field in fields(cls)}
    data_keys = set(data.keys())
    unexpected_fields = data_keys - field_names
    if any(unexpected_fields):
        raise UnexpectedItem(cls, data, unexpected_fields)


def fields_missing_from(data: Mapping, cls: Type[Dataclass]) -> Iterator[Field]:
    """Fields missing from data, but present in the dataclass."""

    def _is_missing(field: Field) -> bool:
        return field.name not in data \
               and field.default is MISSING \
               and field.default_factory is MISSING  # type: ignore # default factory is an unbound function

    return filter(_is_missing, fields(cls))
